<?php

namespace App\Tools;

use App\Key\K;
use EasySwoole\Component\TableManager;
use EasySwoole\Component\Timer;
use EasySwoole\EasySwoole\ServerManager;
use SQLite3;
use Swlib\Saber;
use Swlib\SaberGM;
use Swoole\Coroutine;

class MapDownloader
{
    private $tileUrl = 'https://mt{s}.google.com/vt/lyrs=m&hl=zh-CN&gl=cn&x={x}&y={y}&z={z}';
    /**
     * @var SQLite3
     */
    private $sqlite;

    private $sqliteConfig = [
        'ok' => false,
        'name' => '',
        'sum' => 0,
        'dir' => '',
        'limit' => 50000 * 3,//超过这个值，会重新生成新的sqlite文件
    ];
    private $workerId;
    private $workerHash;
    private $speed = [
        'sum' => 0,//
        'total' => 0,//
        'ok' => 0,
        'has' => 0,
        'commit' => 0,
        'minute' => '',
        'ids' => '',
    ];

    private $thread = [
        'sum' => 10,//开启4个协程来下载数据
        'status' => [],//记录所有协程的状态，如果已经没有了数据，那么，重新拉新的数据
        'done' => 0,//已经有多少个协程下载完成，方便拉新的任务
    ];
    private $work = [
        'url' => [],
        'status' => 0,//这个值是使用来记录，如果服务器没有返回要下载的列表，那么要将最后一个文件上传
    ];

    private $error = [
        'sum' => 0,
        'list' => [],
        'limit' => 40000
    ];

    public function __construct()
    {
        $this->sqliteConfig['dir'] = System::getDownloadMapDbDir();
        if (!is_dir($this->sqliteConfig['dir'])) mkdir($this->sqliteConfig['dir']);
    }

    //如果传入了name 表示设置，否则表示获取数据
    public function workerDbName($name = '')
    {
        $configTable = TableManager::getInstance()->get(K::TABLE_SYSTEM_CONFIG);
        $currentDownloadDbs = $configTable->get(K::CURRENT_MAP_DOWNLOAD_ID);
        if (!$currentDownloadDbs) {
            $currentDownloadDbs = ['number' => [], 'data' => []];
        } else {
            $currentDownloadDbs['data'] = json_decode($currentDownloadDbs['data'], true);
        }
        if ($name) {
            $currentDownloadDbs['data'][$this->workerHash] = $this->sqliteConfig['name'];
            $currentDownloadDbs['data'] = json_encode($currentDownloadDbs['data']);
            $configTable->set(K::CURRENT_MAP_DOWNLOAD_ID, $currentDownloadDbs);
        } else {
            return $currentDownloadDbs['data'];
        }
    }

    public function initSqlite()
    {
        $dbDir = $this->sqliteConfig['dir'];
        $this->sqliteConfig['name'] = 'map=' . date('Y-m-d-H-i-s') . '=' . strtolower(uniqid()) . '=' . mt_rand(1, 1000) . '.db';
        $this->sqliteConfig['file_dir'] = $dbDir . '/' . $this->sqliteConfig['name'];
        $this->sqlite = new SQLite3($dbDir . '/' . $this->sqliteConfig['name']);
        $this->sqlite->query('CREATE TABLE "tile" (
            "id"	TEXT UNIQUE,
            "data"	BLOB,
            PRIMARY KEY("id")
        )');
        //将当前的名称放到table 中，方便在多线程下载的时候进行判断。
        $this->workerDbName($this->sqliteConfig['name']);
        if ($this->sqlite->lastErrorCode()) {
            System::log('create table error', $this->sqlite->lastErrorMsg());
            $this->sqliteConfig['ok'] = false;
            return;
        }
        $this->sqliteConfig['ok'] = true;
//        $this->sqlite->query('PRAGMA temp_store = 2');
    }

    public function workerStart($workerId)
    {
        $this->workerId = $workerId;
        $this->workerHash = 'map-' . $workerId;
        //默认允许两个线程来下载数据
        if ($workerId > 1) return;
        $this->initSqlite();
        if (!$this->sqliteConfig['ok']) {
            System::log('init sqlite error,worker stop');
            return;
        }
        $this->sqlite->query('begin');
        //先开启几个任务来下载数据
        $this->subWorker();
        Timer::getInstance()->loop(30000, [$this, 'noticeError']);
        while (true) {
            $this->getWork();
            Coroutine::sleep(5);
        }
    }

    public function saveDbData($id, $png)
    {
        $sql = 'select id from tile where id=\'' . $id . '\'';
        $has = $this->sqlite->query($sql);
        if ($this->sqlite->lastErrorCode()) {
            echo $this->sqlite->lastErrorMsg(), PHP_EOL;
            $this->error['sum'] += 100;
            $this->error['list'][] = ['id' => $id, 'from' => 'insert-check'];
            return;
        }
        $has = $has->fetchArray(SQLITE3_ASSOC);
        if ($has['id']) {
            $this->speed['has']++;
            return;
        }
        $st = $this->sqlite->prepare('insert into tile (id,data) values (?,?)');
        $st->bindValue('1', $id, SQLITE3_TEXT);
        $st->bindValue('2', $png, SQLITE3_BLOB);
        $st->execute();
        if ($this->sqlite->lastErrorCode()) {
            $this->error['sum'] += 100;
            $this->error['list'][] = ['id' => $id, 'from' => 'insert'];
            echo $this->sqlite->lastErrorMsg(), PHP_EOL;
        }
        $this->speed['ok']++;
    }

    public function getWork()
    {
        $freeSpace = disk_free_space('/');
//        System::log('free space :' . $freeSpace / 1024 / 1024 / 1024);
        if ($freeSpace < 1024 * 1024 * 1024 * 5) {
            System::log('too disk space ' . $freeSpace / 1024 / 1024 / 1024);
            //在没有空间的时候 ，也需要上传数据，这样才将下载的数据删除掉
            $this->noticeDownloadDbs();
            Coroutine::sleep(60);
            return;
        }
//        echo 'count ',count($this->work['url']).' total '.$this->speed['total'],PHP_EOL;
        if (count($this->work['url']) === 0) {
            //当前已经下载完成，那么要判断是不所有的协程也下载完成了
            if ($this->thread['sum'] !== $this->thread['done']) {
                System::log('wait some thread to download ok');
                return;
            }
            //当前已经下载完成，那么需要进行将下载的数据进行提交
            if (!$this->noticeDownloadDbs()) {
                return;
            }
            if (!$this->noticeDownloadIds()) {
                return;
            }
            $speed = $this->speed['sum'] / (microtime(true) - $this->speed['time']);
            $log = ['worker id => ' . $this->workerId, ' speed=>' . $speed, 'ok=>' . $this->speed['ok'], 'has=>' . $this->speed['has'], 'sum=>' . $this->speed['sum'], 'total=>' . $this->speed['total']];
//            print_r($this->speed['ids']);
            $log = implode(' , ', $log);
            System::log($log);
            $this->sqlite->query('commit');
            if ($this->speed['total'] > $this->sqliteConfig['limit'] || ($this->work['status'] == '1' && $this->speed['total'] > 0) || (!System::config(K::ENABLE_MAP_DOWNLOAD) && $this->speed['total'] > 0)) {
                //这里有三个情况需要处理一次：
                //如果在超过了指定的长度，或者在服务器已经没有新任务，同时有下载的数据时,或者是当前不下载街景数据了，那么执行提交操作
                $this->speed['total'] = 0;
                $this->sqlite->close();
                $this->initSqlite();
                System::log('new db');
            }
            $this->sqlite->query('begin');
            $this->speed['ok'] = 0;
            $this->speed['has'] = 0;
            $this->speed['sum'] = 0;
            $this->speed['time'] = microtime(true);
            if (System::checkSpaceOrTrafficIsEnough()) {
                if (System::config(K::ENABLE_MAP_DOWNLOAD)) {
                    $this->createTableData();
                } else {
                    Coroutine::sleep(60);
                }
            } else {
                System::log('traffic is not enough');
            }
        }
    }

    public function createTableData()
    {
        try {
            $data = SaberGM::get(K::MANAGE_SERVER_ADDR_9502 . '/street/map/index', ['timeout' => 5]);
            $all = $data->getBody()->getContents();
            $all = json_decode($all, true);
            //将当前的ID要缓存起来，这样在下载完成 之后，要通知主服务器
            $this->speed['ids'] = $all['data']['ids'];
            foreach ($all['data']['ids'] as $v) {
                $tiles = $this->createTiles($v['id']);
                $this->speed['sum'] += count($tiles);
                foreach ($tiles as $vv) {
                    $this->work['url'][] = $vv;
                }
            }
            if (!$this->speed['sum'] && $all['data']['done']) {
                //如果服务器传递过来的done =1 表示已经没有数据可以下载了，那么要将 work 中的 status 设置成1 ，方便上传最后一个数据库
                $this->work['status'] = 1;
                //同时将系统中可能的错误也清空了
//                $this->error['sum']=0;
            } else {
                $this->work['status'] = 0;
            }
            echo 'current  loop sum,', $this->speed['sum'], PHP_EOL;
        } catch (\Throwable $throwable) {
            System::log('get curl url failed ', $throwable->getMessage());
        }
    }

    public function createTiles($id)
    {
        $r = [];
        $id = explode('$', $id);
        $zoom = current($id);
        $zoom = explode('-', $zoom);
        $zoom = end($zoom);
        $bounds = end($id);
        $bounds = json_decode($bounds, true);
        for ($i = $bounds['s']['x']; $i <= $bounds['e']['x']; $i++) {
            for ($j = $bounds['s']['y']; $j < $bounds['e']['y']; $j++) {
                $r[] = [
                    'id' => $id,
                    'tile_id' => implode('-', [$zoom, $i, $j]),
                    'url' => str_replace(['{s}', '{x}', '{y}', '{z}'], [mt_rand(0, 2), $i, $j, $zoom], $this->tileUrl),
                ];
            }
        }
        return $r;
    }

    public function curl($data)
    {
        try {
            $result = SaberGM::get($data['url'], ['timeout' => 10]);
//            $result = $this->saber->get();
//            $result = SaberGM::get($data['url'], ['timeout' => 20]);
            return $result->getBody()->getContents();
        } catch (\Throwable $throwable) {
            $code = $throwable->getCode();
            echo 'http error code : ', $code, '   ', $throwable->getMessage(), PHP_EOL;
            switch ($code) {
                case -1:
                case -2:
                case -4:
                case -3:
                    $this->work['url'][] = $data;
                    break;
                case 404:
                case 400:
                    $this->error['list'][] = ['id' => $data['tile_id'], 'data' => $data, 'from' => $code];
//                    foreach ($this->work['url'] as $k => $v) {
//                        if (strpos($v['url'], $data['id']) !== false) unset($this->work['url'][$k]);
//                    }
                    break;
                default:
                    $this->error['list'][] = ['id' => $data['tile_id'], 'data' => $data, 'from' => 'download', 'code' => $code];
                    echo 'worker_id ', $this->workerId, ' === ', $throwable->getMessage(), PHP_EOL;
                    $this->error['sum']++;
            }
        }
        return false;
    }

    public function subWorker()
    {
        for ($i = 0; $i < $this->thread['sum']; $i++) {
            go([$this, 'download'], $i);
        }
    }

    public function download($tag)
    {
        while (true) {
            if ($this->error['sum'] > $this->error['limit']) {
                echo 'too many errors', PHP_EOL;
                $this->sqlite->query('commit');
                return;
            } elseif (!$this->sqliteConfig['ok']) {
                System::log('sqlite is error,so stop download');
                return;
            } else {
                $size = count($this->work['url']);
                if ($size && $size % 400 === 0) {
                    echo 'worker id ', $this->workerId, ' count ', $size, ' error', $this->error['sum'], PHP_EOL;
                }
                $data = array_shift($this->work['url']);
                if ($data) {
                    $this->thread['status'][$tag] = false;
                    $png = $this->curl($data);
                    if ($png) {
                        //这里使用total 来表示下载了多少次，方便后面进行事务的提交
                        $this->speed['total']++;
                        $this->saveDbData($data['tile_id'], $png);
                        if ($this->speed['total'] && $this->speed['total'] % 200 === 0) {
                            $this->sqlite->query('commit');
                            $this->sqlite->query('begin');
                        }
                    }
                } else {
                    //已经没有数据了，那么要判断是不是所有的下载进程已经下载完成了
                    $this->thread['status'][$tag] = true;
                    $done = 0;
                    for ($i = 0; $i < $this->thread['sum']; $i++) {
                        if ($this->thread['status'][$i]) {
                            $done++;
                        }
                    }
                    $this->thread['done'] = $done;
                    Coroutine::sleep(1);
                }
            }
        }
    }

    private function noticeDownloadIds()
    {
        if (!$this->speed['ids']) return true;
        $result = false;
        try {
            SaberGM::post(K::MANAGE_SERVER_ADDR_9502 . '/street/map/confirm', ['ids' => $this->speed['ids']], ['timeout' => 5]);
            $result = true;
        } catch (\Throwable $throwable) {
            System::log('upload download ids error', $throwable->getMessage());
        }
        return $result;
    }

    public function noticeError()
    {
        if (!$this->error['list']) return;
        $post = $this->error['list'];
        try {
            //因为使用了定时器，防止在提交的过程，将新产生的数据丢弃
            $this->error['list'] = [];
            SaberGM::post(K::MANAGE_SERVER_ADDR_9502 . '/street/map/error', $post, ['timeout' => 5]);
        } catch (\Throwable $throwable) {
            $this->error['list'] = array_merge($post, $this->error['list']);
            System::log('upload error list error ' . $throwable->getMessage());
        }
    }

    private function noticeDownloadDbs()
    {
        $result = false;
        try {
            $dbs = scandir($this->sqliteConfig['dir']);
            $dbs = array_filter($dbs, function ($v) {
                return strtolower(pathinfo($v, PATHINFO_EXTENSION)) === 'db';
            });
            $workerDb = $this->workerDbName();
            echo 'worker dbs', PHP_EOL;
            print_r($workerDb);
            foreach ($dbs as $k => $v) {
                if ($v === $this->sqliteConfig['name'] || in_array($v, $workerDb)) unset($dbs[$k]);
            }
            $dbs = array_values($dbs);
            if (!$dbs) return true;
            $response = SaberGM::post(K::MANAGE_SERVER_ADDR_9502 . '/street/map/db', ['dbs' => $dbs], ['timeout' => 5]);
            $str = $response->getBody()->getContents();
            $str = json_decode($str, true);
            $usedTraffic = System::config(K::STREET_TRAFFIC_USED);
            foreach ($str['data']['del'] as $d) {
                //要删除的数据,如果是当前正在使用的，或者是别的线程在下载的，那么不能删除
                if ($d === $this->sqliteConfig['name'] || in_array($d, $workerDb)) continue;
                $usedTraffic = $usedTraffic + filesize($this->sqliteConfig['dir'] . '/' . $d);
                $un = unlink($this->sqliteConfig['dir'] . '/' . $d) ? 'success' : 'failed';
                System::log('unlink ' . $d . ' status: ' . $un);
            }
            System::config(K::STREET_TRAFFIC_USED, $usedTraffic);
            $result = true;
        } catch (\Throwable $throwable) {
            System::log('upload download dbs error', $throwable->getMessage());
        }
        return $result;
    }
}
